#ifndef STATIC_CONTAINER_LODGE_LIST_H

#define STATIC_CONTAINER_LODGE_LIST_H

#include "static_container/destruct.h"
#include "static_container/list_node_pool.h"
#include "static_container/STATIC_CONTAINER_MEMBERTYPEDEF.h"
#include "static_container/compare_methods.h"
#include <algorithm>
#include <functional>
#include <boost/call_traits.hpp>
#include <boost/assert.hpp>

namespace static_container {
	///	std::list ݊̊Ԏ؂胊Xg
	/**
	 *	list_node_pool m[h؂ăNXgȂĂB
	 *	pIAlist_node_pool Ƀm[hԂB
	 *
	 *	lodge_list p list_node_pool jĂ͂ȂȂB
	 *
	 *	http://www.alc.co.jp/ ŁA
	 *	lodge = y-1z ܂Ahoキ͂pAhoキp
	 *	Ƃ̂ŁAAƗǂO΁AWB
	 */
	template < typename Value >
	class lodge_list : public compare_methods< lodge_list< Value > > {
	public:
		typedef list_link			link;
		typedef list_node< Value >	node;
		typedef abstruct_list_node_pool<
			Value >					pool;
		STATIC_CONTAINER_MEMBERTYPEDEF( Value )
		typedef typename boost::call_traits< Value >::param_type	param_type;
	private:
		static void destruct( Value& v ) {
			static_container::destruct< Value >( v );
		}

		class iterator_base {
		protected:
			link*	node_;
			void setNext( link* next ) {
				BOOST_ASSERT( 0 != node_ );
				node_->next = next;
			}
			void setPrev( link* prev ) {
				BOOST_ASSERT( 0 != node_ );
				node_->prev = prev;
			}
			link* getNode() {
				BOOST_ASSERT( 0 != node_ );
				return node_;
			}
			friend class lodge_list;

		public:
			bool equal( iterator_base const& other ) const {
				return this->node_ == other.node_;
			}
			void increment() {
				BOOST_ASSERT( 0 != node_ );
				node_ = node_->next;
			}
			void decrement() {
				BOOST_ASSERT( 0 != node_ );
				node_ = node_->prev;
			}
		};
	public:
		class iterator;
		
		///	萔QƑoCe[^[
		class const_iterator :
			public iterator_base,
			public std::iterator< std::bidirectional_iterator_tag, Value const > {

			friend class lodge_list;
			friend class iterator;
			explicit const_iterator( const link* n ) {
				node_ = const_cast< link* >( n );
			}
		public:
			const_iterator() { node_ = 0; }
			const_reference operator * () const {
				BOOST_ASSERT( 0 != node_ );
				return static_cast< const node* >( node_ )->value;
			}
			const_pointer operator -> () const {
				BOOST_ASSERT( 0 != node_ );
				return &static_cast< const node* >( node_ )->value;
			}
			const_iterator& operator ++ () {
				increment();
				return *this;
			}
			const_iterator& operator -- () {
				decrement();
				return *this;
			}
			const_iterator operator ++ ( int ) {
				const_iterator result( *this );
				operator ++ ();
				return result;
			}
			const_iterator operator -- ( int ) {
				const_iterator result( *this );
				operator -- ();
				return result;
			}
			bool operator == (const const_iterator& x) const {
				return equal( x );
			}
			bool operator != (const const_iterator& x) const {
				return !equal( x );
			}
		};

		///	oCe[^[
		class iterator :
			public iterator_base,
			public std::iterator< std::bidirectional_iterator_tag, Value > {

			friend class lodge_list;
			explicit iterator( link* n ) {
				node_ = n;
			}
		public:
			operator const_iterator () {
				return const_iterator( node_ );
			}
			iterator() { node_ = 0; }
			reference operator * () const {
				BOOST_ASSERT( 0 != node_ );
				return static_cast< node* >( node_ )->value;
			}
			Value* operator -> () {
				BOOST_ASSERT( 0 != node_ );
				return &static_cast< node* >( node_ )->value;
			}
			iterator& operator ++ () {
				increment();
				return *this;
			}
			iterator& operator -- () {
				decrement();
				return *this;
			}
			iterator operator ++ ( int ) {
				iterator result( *this );
				operator ++ ();
				return result;
			}
			iterator operator -- ( int ) {
				iterator result( *this );
				operator -- ();
				return result;
			}
			bool operator == (const iterator& x) const {
				return equal( x );
			}
			bool operator != (const iterator& x) const {
				return !equal( x );
			}
		};

	private:
		link	end_;
		pool*	pool_;

		void init() {
			end_.next = &end_;
			end_.prev = &end_;
		}

	public:
		lodge_list( pool& ioPool ) : pool_( &ioPool ) {
			init();
		}
		lodge_list( const lodge_list& other ) : pool_( other.pool_ ) {
			init();
			insert( begin(), other.begin(), other.end() );			
		}
		~lodge_list() {
			clear();
		}
		
		lodge_list& operator = ( const lodge_list& other ) {
			if ( this != &other ) {
				clear();
				pool_ = other.pool_;
				insert( begin(), other.begin(), other.end() );
			}
			return *this;
		}

		iterator begin() {
			return iterator( end_.next );
		}
		iterator end() {
			return iterator( &end_ );
		}
		const_iterator begin() const {
			return const_iterator( end_.next );
		}
		const_iterator end() const {
			return const_iterator( &end_ );
		}

		reference push_back() {
			node* n = pool_->allocate();
			BOOST_ASSERT( 0 != n );
			new( &n->value ) Value();
		}

		///	m[hv[擾AɘA邪AValue ̃RXgN^͌ĂяoȂ
		/**
		 *	m[h́Apos ̒OɘA
		 */
		Value* allocate( iterator pos = end() ) {
			node* n = pool_->allocate();
			BOOST_ASSERT( 0 != n );
			iterator prev = pos;
			--prev;
			prev.setNext( n );
			pos.setPrev( n );
			n->next = pos.getNode();
			n->prev = prev.getNode();
			return &n->value;
		}

		///	pos ʒuɂvf̒O v }
		/**
		 *	prev = pos - 1 ƂƁAX prev, pos ƐڑĂ̂Aprev, v, pos ̏ɐڑB
		 */
		void insert( iterator pos, param_type v ) {
			Value* val = allocate( pos );
			//	RXgN^Ăяo
			new( val ) Value( v );
		}

		///	pos ʒuɂvf̒O v }
		/**
		 *	prev = pos - 1 ƂƁAX prev, pos ƐڑĂ̂Aprev, v, pos ̏ɐڑB
		 */
		template < typename It >
		void insert( iterator pos, It first, It last ) {
			for ( ; first != last; ++first ) {
				Value* val = allocate( pos );
				//	RXgN^Ăяo
				new( val ) Value( *first );
			}
		}
		
		///	擪ɒǉ
		void push_front( param_type v ) {
			return insert( begin(), v );
		}

		///	ɒǉ
		void push_back( param_type v ) {
			return insert( end(), v );
		}

		///	pos Ŏvf폜
		void erase( iterator pos ) {
			//	
			destruct( *pos );

			//	node  pool ɕԂ
			pool_->deallocate( pos.getNode() );
		}

		///	 [ first, last ) 폜
		void erase( iterator first, iterator last ) {
			//	
			std::for_each( first, last, destruct );
			//	node  pool ɕԂ
			pool_->deallocate( first.getNode(), last.getNode() );
		}

		///	NAiSvf폜j
		void clear() {
			erase( begin(), end() );
		}

		///	擪폜
		void pop_front() {
			erase( begin() );
		}

		///	폜
		void pop_back() {
			erase( --end() );
		}

		///	l value vf폜
		void remove( const Value& value ) {
			remove_if( std::bind2nd( std::equal_to< value_type >(), value ) );
		}

		///	t폜
		/**
		 *	pred( *it )  true ԂƂ it 폜
		 */
		template < typename Pred >
		void remove_if( Pred pred ) {
			for ( iterator it = begin(); end() != it; ) {
				if ( pred( *it ) ) {
					erase( it++ );
				} else {
					++it;
				}
			}
		}

		///	TCY擾
		size_type size() const {
			return static_cast< size_type >( std::distance( begin(), end() ) );
		}

		///	󂩂ǂH
		bool empty() const {
			return end_.next == &end_;
		}

		///	擪擾
		reference front() {
			return *begin();
		}
		///	擪擾
		param_type front() const {
			return *begin();
		}
		///	擾
		reference back() {
			return *( --end() );
		}
		///	擾
		param_type back() const {
			return *( --end() );
		}
	};
}

#endif