#ifndef STATIC_CONTAINER_STRING_H

#define STATIC_CONTAINER_STRING_H

#include <iosfwd>
#include <boost/assert.hpp>
#include "static_container/STATIC_CONTAINER_MEMBERTYPEDEF.h"
#include "static_container/integer.h"
#include <algorithm>

namespace static_container {
	///	ÓI
	/**
	 *	̔zȂ̂ŁAoCit@Cւ̖ߍ݂ȒPłB
	 *
	 *	Iɂ́ACh[ MaxStrLen + 1 ] ݂̂𗘗pB
	 */
	template< size_type MaxStrLen, typename Ch = char, typename ChTraits = std::char_traits< Ch > >
	class string {
	public:
		STATIC_CONTAINER_MEMBERTYPEDEF( Ch )
		typedef Ch				char_type;
		typedef ChTraits		traits_type;
		typedef pointer			iterator;
		typedef const_pointer	const_iterator;
	private:
		Ch	buffer_[ MaxStrLen + 1 ];
	public:
		///	őTCY
		BOOST_STATIC_CONSTANT( size_type, npos = -1 );
		BOOST_STATIC_CONSTANT( size_type, const_max = MaxStrLen );

		///	őTCY擾
		static size_type max_size() { return const_max; }

		///	eʎ擾
		static size_type capaciry() { return const_max; }
	
		/// TCY擾
		size_type size() const { return traits_type::length( buffer_ ); }

		/// TCY擾
		size_type length() const { return size(); }

		iterator begin() { return buffer_; }
		const_iterator begin() const { return buffer_; }
		iterator end() { return begin() + size(); }
		const_iterator end() const { return begin() + size(); }

		char_type& operator [] ( size_type i ) {
			PYD_ASSERT( i < size() );
			return *( begin() + i );
		}
		const char_type& operator [] ( size_type i ) const {
			PYD_ASSERT( i < size() );
			return *( begin() + i );
		}
		bool empty() const {
			return 0 == size();
		}
		reference at( size_type i ) { return operator [] ( i ); }
		char_type at( size_type i ) const { return operator [] ( i ); }

		reference front() { return *begin(); }
		char_type front() const { return *begin(); }
		reference back() { return *( end() - 1 ); }
		char_type back() const { return *( end() - 1 ); }
	
		///	ꕶǉ
		void push_back( char_type ch ) {
			BOOST_ASSERT( size() < MaxStrLen );
			iterator last = end();
			*last = ch;
			++last;
			*last = char_type();
		}

		///	ꕶǉ
		void append( char_type ch ) {
			push_back( ch );
		}

		///	𖖔ɒǉ
		void append( const char* str ) {
			BOOST_ASSERT( 0 != str && size() + traits_type::length( str ) <= max_size() );
			traits_type::copy( end(), str, traits_type::length( str ) + 1 );
		}

		///	𖖔ɒǉ
		template < size_type OtherMaxStrLen >
		void append( const string< OtherMaxStrLen, Ch, ChTraits >& other ) {
			append( other.c_str() );
		}

		///	ꕶ폜
		void pop_back() {
			BOOST_ASSERT( false == empty() );
			iterator last = end();
			--last;
			*last = char_type();
		}

		/// ftHgRXgN^
		string() {
			clear();
		}
		
		/// Rs[RXgN^
		template < size_type OtherMaxStrLen >
		string( const string< OtherMaxStrLen, Ch, ChTraits >& other ) {
			if ( buffer_ != other.begin() ) {
				BOOST_ASSERT( other.size() <= max_size() );
				traits_type::copy( buffer_, other.begin(), other.size() + 1 );
			}
		}

		string( const char_type* s ) {
			if ( s != buffer_ ) {
				BOOST_ASSERT( traits_type::length( s ) <= max_size() );
				traits_type::copy( buffer_, s, traits_type::length( s ) + 1 );
			}
		}

		// fXgN^
		~string() {
		}

		/// Zq
		string& operator = ( const string& other )	{
			return operator = ( other.c_str() );
		}
		string& operator = ( const char_type* s ) {
			if ( s != buffer_ ) {
				BOOST_ASSERT( traits_type::length( s ) <= max_size() );
				traits_type::copy( buffer_, s, traits_type::length( s ) + 1 );
			}
			return *this;
		}
		
		/// ANZX
		const_pointer c_str() const { return begin(); }

		///	append() Ɠł
		string operator += ( const char* other ) {
			append( other );
			return *this;
		}

		///	append() Ɠł
		template < size_type OtherMaxStrLen >
		string operator += ( const string< OtherMaxStrLen, Ch, ChTraits >& other ) {
			append( other );
			return *this;
		}
		
		///	󕶎
		void clear() {
			buffer_[ 0 ] = char_type();
		}

		///	H
		template < size_type OtherMaxStrLen >
		friend bool operator == ( const string& a, const string< OtherMaxStrLen, Ch, ChTraits >& b ) {
			return operator == ( a, b.c_str() );
		}

		///	H
		friend bool operator == ( const string& a, const char* b ) {
			size_type size = a.size();
			return char_type() == b[ size ] &&
				std::equal( a.begin(), a.begin() + size, b, traits_type::eq );
		}

		///	H
		friend bool operator == ( const char* a, const string& b ) {
			return operator == ( b, a );
		}
		
		///	قȂH
		template < size_type OtherMaxStrLen >
		friend bool operator != ( const string& a, const string< OtherMaxStrLen, Ch, ChTraits >& b ) {
			return !operator == ( a, b );
		}

		///	قȂH
		friend bool operator != ( const string& a, const char* b ) {
			return !operator == ( a, b );
		}

		///	قȂH
		friend bool operator != ( const char* a, const string& b ) {
			return !operator == ( a, b );
		}

		///	r
		template < size_type OtherMaxStrLen >
		friend bool operator < ( const string& a, const string< OtherMaxStrLen, Ch, ChTraits >& b ) {
			return operator < ( a, b.c_str() );
		}

		///	r
		friend bool operator < ( const string& a, const char* b ) {
			return std::lexicographical_compare(
				a.begin(), a.end(),
				b, b + traits_type::length( b ),
				traits_type::eq_int_type );
		}

		///	r
		friend bool operator < ( const char* a, const string& b ) {
			return operator < ( b, a );
		}
	};
}

#endif